@@ -1,5 +1,5 @@
 /* Job execution and handling for GNU Make.
-Copyright (C) 1988, 89, 90, 91, 92, 93, 94 Free Software Foundation, Inc.
+Copyright (C) 1988, 89, 90, 91, 92, 93, 94, 1995 Free Software Foundation, Inc.
 This file is part of GNU Make.
 
 GNU Make is free software; you can redistribute it and/or modify
@@ -23,9 +23,6 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 #include "variable.h"
 #include <assert.h>
 
-/* Default path to search for executables.  */
-static char default_path[] = ":/bin:/usr/bin";
-
 /* Default shell to use.  */
 char default_shell[] = "/bin/sh";
 
@@ -125,6 +122,9 @@ extern int setgid (), getgid ();
 #ifdef HAVE_GETDTABLESIZE
 extern int getdtablesize ();
 #else
+#ifdef HAVE_SYSCONF_OPEN_MAX
+#define getdtablesize()		((int) sysconf (_SC_OPEN_MAX))
+#else
 #include <sys/param.h>
 #define getdtablesize() NOFILE
 #if !defined (NOFILE) && defined (NOFILES_MAX)
@@ -133,6 +133,7 @@ extern int getdtablesize ();
 #endif
 #endif
 #endif
+#endif
 
 extern int getloadavg ();
 extern int start_remote_job_p ();
@@ -173,14 +174,9 @@ child_error (target_name, exit_code, exit_sig, coredump, ignored)
 	   "*** [%s] Error %d",
 	   target_name, exit_code);
   else
-    {
-      char *coredump_string = coredump ? " (core dumped)" : "";
-      if (exit_sig > 0 && exit_sig < NSIG)
-	error ("*** [%s] %s%s",
-	       target_name, sys_siglist[exit_sig], coredump_string);
-      else
-	error ("*** [%s] Signal %d%s", target_name, exit_sig, coredump_string);
-    }
+    error ("*** [%s] %s%s",
+	   target_name, strsignal (exit_sig),
+	   coredump ? " (core dumped)" : "");
 }
 
 static unsigned int dead_children = 0;
@@ -381,7 +377,7 @@ reap_children (block, err)
 		 delete non-precious targets, and abort.  */
 	      static int delete_on_error = -1;
 	      child_error (c->file->name, exit_code, exit_sig, coredump, 0);
-	      c->file->update_status = 1;
+	      c->file->update_status = 2;
 	      if (delete_on_error == -1)
 		{
 		  struct file *f = lookup_file (".DELETE_ON_ERROR");
@@ -409,7 +405,7 @@ reap_children (block, err)
 			 Since there are more commands that wanted to be run,
 			 the target was not completely remade.  So we treat
 			 this as if a command had failed.  */
-		      c->file->update_status = 1;
+		      c->file->update_status = 2;
 		    }
 		  else
 		    {
@@ -437,7 +433,7 @@ reap_children (block, err)
 	    }
 
 	  /* When we get here, all the commands for C->file are finished
-	     (or aborted) and C->file->update_status contains 0 or 1.  But
+	     (or aborted) and C->file->update_status contains 0 or 2.  But
 	     C->file->command_state is still cs_running if all the commands
 	     ran; notice_finish_file looks for cs_running to tell it that
 	     it's interesting to check the file's modtime again now.  */
@@ -591,6 +587,7 @@ start_job_command (child)
 
   if (argv == 0)
     {
+    next_command:
       /* This line has no commands.  Go to the next.  */
       if (job_next_command (child))
 	start_job_command (child);
@@ -623,10 +620,7 @@ start_job_command (child)
     {
       free (argv[0]);
       free ((char *) argv);
-      if (job_next_command (child))
-	start_job_command (child);
-      child->file->update_status = 0;
-      return;
+      goto next_command;
     }
 
   /* Flush the output streams so they won't have things written twice.  */
@@ -689,6 +683,8 @@ start_job_command (child)
     {
       /* Fork the child process.  */
 
+      char **parent_environ;
+
 #ifdef	 POSIX
       (void) sigprocmask (SIG_BLOCK, &fatal_signal_set, (sigset_t *) 0);
 #else
@@ -698,7 +694,9 @@ start_job_command (child)
 #endif
 
       child->remote = 0;
+      parent_environ = environ;
       child->pid = vfork ();
+      environ = parent_environ;	/* Restore value child may have clobbered.  */
       if (child->pid == 0)
 	{
 	  /* We are the child side.  */
@@ -768,7 +766,7 @@ start_waiting_job (c)
     {
       /* Put this child on the chain of children waiting
 	 for the load average to go down.  */
-      c->file->command_state = cs_running;
+      set_command_state (c->file, cs_running);
       c->next = waiting_jobs;
       waiting_jobs = c;
       return 0;
@@ -1073,105 +1071,6 @@ child_execute_job (stdin_fd, stdout_fd, argv, envp)
   exec_command (argv, envp);
 }
 
-/* Search PATH for FILE.
-   If successful, store the full pathname in PROGRAM and return 1.
-   If not sucessful, return zero.  */
-
-static int
-search_path (file, path, program)
-     char *file, *path, *program;
-{
-  if (path == 0 || path[0] == '\0')
-    path = default_path;
-
-  if (
-#ifdef __MSDOS__
-      strpbrk (file, "/\\:")
-#else
-      index (file, '/') 
-#endif
-      != 0)
-    {
-      strcpy (program, file);
-      return 1;
-    }
-  else
-    {
-      unsigned int len;
-
-#ifdef	HAVE_GETGROUPS
-#ifndef	HAVE_UNISTD_H
-      extern int getgroups ();
-#endif
-      static int ngroups = -1;
-#ifdef	NGROUPS_MAX
-      static GETGROUPS_T groups[NGROUPS_MAX];
-#define	ngroups_max	NGROUPS_MAX
-#else
-      static GETGROUPS_T *groups = 0;
-      static int ngroups_max;
-      if (groups == 0)
-	{
-	  ngroups_max = GET_NGROUPS_MAX;
-	  groups = (GETGROUPS_T *) malloc (ngroups_max * sizeof (GETGROUPS_T));
-	}
-#endif
-      if (groups != 0 && ngroups == -1)
-	ngroups = getgroups (ngroups_max, groups);
-#endif	/* Have getgroups.  */
-
-      len = strlen (file) + 1;
-      do
-	{
-	  struct stat st;
-	  int perm;
-	  char *p;
-
-	  p = index (path, PATH_SEPARATOR_CHAR);
-	  if (p == 0)
-	    p = path + strlen (path);
-
-	  if (p == path)
-	    bcopy (file, program, len);
-	  else
-	    {
-	      bcopy (path, program, p - path);
-	      program[p - path] = '/';
-	      bcopy (file, program + (p - path) + 1, len);
-	    }
-
-	  if (safe_stat (program, &st) == 0
-	      && S_ISREG (st.st_mode))
-	    {
-	      if (st.st_uid == geteuid ())
-		perm = (st.st_mode & 0100);
-	      else if (st.st_gid == getegid ())
-		perm = (st.st_mode & 0010);
-	      else
-		{
-#ifdef	HAVE_GETGROUPS
-		  register int i;
-		  for (i = 0; i < ngroups; ++i)
-		    if (groups[i] == st.st_gid)
-		      break;
-		  if (i < ngroups)
-		    perm = (st.st_mode & 0010);
-		  else
-#endif	/* Have getgroups.  */
-		    perm = (st.st_mode & 0001);
-		}
-
-	      if (perm != 0)
-		return 1;
-	    }
-
-	  path = p + 1;
-	} while (*path != '\0');
-    }
-
-  return 0;
-}
-
 /* Replace the current process with one running the command in ARGV,
    with environment ENVP.  This function does not return.  */
 
@@ -1179,82 +1078,65 @@ void
 exec_command (argv, envp)
      char **argv, **envp;
 {
-  char *shell, *path;
-  PATH_VAR (program);
-  register char **ep;
-
-  shell = path = 0;
-  for (ep = envp; *ep != 0; ++ep)
-    {
-      if (shell == 0 && !strncmp(*ep, "SHELL=", 6))
-	shell = &(*ep)[6];
-      else if (path == 0 && !strncmp(*ep, "PATH=", 5))
-	path = &(*ep)[5];
-      else if (path != 0 && shell != 0)
-	break;
-    }
-
   /* Be the user, permanently.  */
   child_access ();
 
-  if (!search_path (argv[0], path, program))
-    error ("%s: Command not found", argv[0]);
-  else
+  /* Run the program.  */
+  environ = envp;
+  execvp (argv[0], argv);
+  
+  switch (errno)
     {
-      /* Run the program.  */
-      execve (program, argv, envp);
+    case ENOENT:
+      error ("%s: Command not found", argv[0]);
+      break;
+    case ENOEXEC:
+      {
+	/* The file is not executable.  Try it as a shell script.  */
+	extern char *getenv ();
+	char *shell;
+	char **new_argv;
+	int argc;
 
-      if (errno == ENOEXEC)
-	{
-	  PATH_VAR (shell_program);
-	  char *shell_path;
-	  if (shell == 0)
-	    shell_path = default_shell;
-	  else
-	    {
-	      if (search_path (shell, path, shell_program))
-		shell_path = shell_program;
-	      else
-		{
-		  shell_path = 0;
-		  error ("%s: Shell program not found", shell);
-		}
-	    }
+	shell = getenv ("SHELL");
+	if (shell == 0)
+	  shell = default_shell;
 
-	  if (shell_path != 0)
-	    {
-	      char **new_argv;
-	      int argc;
+	argc = 1;
+	while (argv[argc] != 0)
+	  ++argc;
 
-	      argc = 1;
-	      while (argv[argc] != 0)
-		++argc;
+	new_argv = (char **) alloca ((1 + argc + 1) * sizeof (char *));
+	new_argv[0] = shell;
+	new_argv[1] = program;
+	while (argc > 0)
+	  {
+	    new_argv[1 + argc] = argv[argc];
+	    --argc;
+	  }
 
-	      new_argv = (char **) alloca ((1 + argc + 1) * sizeof (char *));
-	      new_argv[0] = shell_path;
-	      new_argv[1] = program;
-	      while (argc > 0)
-		{
-		  new_argv[1 + argc] = argv[argc];
-		  --argc;
-		}
+	execvp (shell, new_argv);
+	if (errno == ENOENT)
+	  error ("%s: Shell program not found", shell);
+	else
+	  perror_with_name ("execvp: ", shell);
+	break;
+      }
 
-	      execve (shell_path, new_argv, envp);
-	      perror_with_name ("execve: ", shell_path);
-	    }
-	}
-      else
-	perror_with_name ("execve: ", program);
+    default:
+      perror_with_name ("execvp: ", argv[0]);
+      break;
     }
 
   _exit (127);
 }
 
-/* Figure out the argument list necessary to run LINE as a command.
-   Try to avoid using a shell.  This routine handles only ' quoting.
-   Starting quotes may be escaped with a backslash.  If any of the
-   characters in sh_chars[] is seen, or any of the builtin commands
-   listed in sh_cmds[] is the first word of a line, the shell is used.
+/* Figure out the argument list necessary to run LINE as a command.  Try to
+   avoid using a shell.  This routine handles only ' quoting, and " quoting
+   when no backslash, $ or ` characters are seen in the quotes.  Starting
+   quotes may be escaped with a backslash.  If any of the characters in
+   sh_chars[] is seen, or any of the builtin commands listed in sh_cmds[]
+   is the first word of a line, the shell is used.
 
    If RESTP is not NULL, *RESTP is set to point to the first newline in LINE.
    If *RESTP is NULL, newlines will be ignored.
@@ -1332,7 +1214,7 @@ construct_command_argv_internal (line, restp, shell, ifs)
 	string_char:
 	  /* Inside a string, just copy any char except a closing quote
 	     or a backslash-newline combination.  */
-	  if (*p == '\'')
+	  if (*p == instring)
 	    instring = 0;
 	  else if (*p == '\\' && p[1] == '\n')
 	    goto swallow_escaped_newline;
@@ -1342,6 +1224,10 @@ construct_command_argv_internal (line, restp, shell, ifs)
 	      *restp = p;
 	      goto end_of_line;
 	    }
+	  /* Backslash, $, and ` are special inside double quotes.
+	     If we see any of those, punt.  */
+	  else if (instring == '"' && index ("\\$`", *p) != 0)
+	    goto slow;
 	  else
 	    *ap++ = *p;
 	}
@@ -1399,7 +1285,8 @@ construct_command_argv_internal (line, restp, shell, ifs)
 	    break;
 
 	  case '\'':
-	    instring = 1;
+	  case '"':
+	    instring = *p;
 	    break;
 
 	  case '\n':
@@ -1505,7 +1392,7 @@ construct_command_argv_internal (line, restp, shell, ifs)
 	 strcat (dos_bname, ".bat");
 	 strcat (dos_bename, ".err");
        }
-     batch = fopen(bename, "w"); /* Create a file.  */
+     batch = fopen (dos_bename, "w"); /* Create a file.  */
      if (batch != NULL)
        fclose (batch);
      batch = fopen (dos_bname, "w");
@@ -1514,7 +1401,7 @@ construct_command_argv_internal (line, restp, shell, ifs)
      fprintf (batch, "\nif errorlevel 1 del %s\n", dos_bename);
      fclose (batch);
      new_argv = (char **) xmalloc(2 * sizeof(char *));
-     new_argv[0] = strdup (bname);
+     new_argv[0] = strdup (dos_bname);
      new_argv[1] = 0;
    }
 #else	/* Not MSDOS.  */
@@ -1564,7 +1451,7 @@ construct_command_argv_internal (line, restp, shell, ifs)
 	    continue;
 	  }
 
-	if (*p == '\\' || *p == '\''
+	if (*p == '\\' || *p == '\'' || *p == '"'
 	    || isspace (*p)
 	    || index (sh_chars, *p) != 0)
 	  *ap++ = '\\';
@@ -1580,11 +1467,12 @@ construct_command_argv_internal (line, restp, shell, ifs)
   return new_argv;
 }
 
-/* Figure out the argument list necessary to run LINE as a command.
-   Try to avoid using a shell.  This routine handles only ' quoting.
-   Starting quotes may be escaped with a backslash.  If any of the
-   characters in sh_chars[] is seen, or any of the builtin commands
-   listed in sh_cmds[] is the first word of a line, the shell is used.
+/* Figure out the argument list necessary to run LINE as a command.  Try to
+   avoid using a shell.  This routine handles only ' quoting, and " quoting
+   when no backslash, $ or ` characters are seen in the quotes.  Starting
+   quotes may be escaped with a backslash.  If any of the characters in
+   sh_chars[] is seen, or any of the builtin commands listed in sh_cmds[]
+   is the first word of a line, the shell is used.
 
    If RESTP is not NULL, *RESTP is set to point to the first newline in LINE.
    If *RESTP is NULL, newlines will be ignored.
