@@ -1,5 +1,5 @@
 /* Basic dependency engine for GNU Make.
-Copyright (C) 1988, 89, 90, 91, 92, 93, 94 Free Software Foundation, Inc.
+Copyright (C) 1988, 89, 90, 91, 92, 93, 94, 1995 Free Software Foundation, Inc.
 This file is part of GNU Make.
 
 GNU Make is free software; you can redistribute it and/or modify
@@ -491,7 +491,16 @@ update_file_1 (file, depth)
       return dep_status;
     }
 
-  set_command_state (file, cs_not_started);
+  if (file->command_state == cs_deps_running)
+    /* The commands for some deps were running on the last iteration, but
+       they have finished now.  Reset the command_state to not_started to
+       simplify later bookkeeping.  It is important that we do this only
+       when the prior state was cs_deps_running, because that prior state
+       was definitely propagated to FILE's also_make's by set_command_state
+       (called above), but in another state an also_make may have
+       independently changed to finished state, and we would confuse that
+       file's bookkeeping (updated, but not_started is bogus state).  */
+    set_command_state (file, cs_not_started);
 
   /* Now record which dependencies are more
      recent than this file, so we can define $?.  */
@@ -547,7 +556,6 @@ update_file_1 (file, depth)
   if (!must_make)
     {
       DEBUGPR ("No need to remake target `%s'.\n");
-      file->update_status = 0;
       notice_finished_file (file);
       return 0;
     }
@@ -565,14 +573,17 @@ update_file_1 (file, depth)
 
   switch (file->update_status)
     {
-    case 1:
+    case 2:
       DEBUGPR ("Failed to remake target file `%s'.\n");
       break;
     case 0:
       DEBUGPR ("Successfully remade target file `%s'.\n");
       break;
+    case 1:
+      DEBUGPR ("Target file `%s' needs remade under -q.\n");
+      break;
     default:
-      assert (file->update_status == 0 || file->update_status == 1);
+      assert (file->update_status >= 0 && file->update_status <= 2);
       break;
     }
 
@@ -596,12 +607,12 @@ notice_finished_file (file)
 
   if (touch_flag
       /* The update status will be:
-	 	-1	if no commands were run;
-		0	if some commands (+ or ${MAKE}) were run and won;
+	 	-1	if this target was not remade;
+		0	if 0 or more commands (+ or ${MAKE}) were run and won;
 		1	if some commands were run and lost.
-	 The only time we don't want to touch the target is if
-	 it had some recursive commands, and they lost.  */
-      && file->update_status != 1)
+	 We touch the target if it has commands which either were not run
+	 or won when they ran (i.e. status is 0).  */
+      && file->update_status == 0)
     {
       if (file->cmds != 0 && file->cmds->any_recurse)
 	{
@@ -639,7 +650,7 @@ notice_finished_file (file)
 	f->last_mtime = file->last_mtime;
     }
 
-  if (file->update_status != -1)
+  if (ran && file->update_status != -1)
     /* We actually tried to update FILE, which has
        updated its also_make's as well (if it worked).
        If it didn't work, it wouldn't work again for them.
@@ -825,15 +836,29 @@ remake_file (file)
       else
 	{
 	  /* This is a dependency file we cannot remake.  Fail.  */
-	  static char noway[] = "No rule to make target";
+	  static const char msg_noparent[]
+	    = "%sNo rule to make target `%s'%s";
+	  static const char msg_parent[]
+	    = "%sNo rule to make target `%s', needed by `%s'%s";
 	  if (keep_going_flag || file->dontcare)
 	    {
 	      if (!file->dontcare)
-		error ("*** %s `%s'.", noway, file->name);
- 	      file->update_status = 1;
+		{
+		  if (file->parent == 0)
+		    error (msg_noparent, "*** ", file->name, ".");
+		  else
+		    error (msg_parent, "*** ",
+			   file->name, file->parent->name, ".");
+		}
+ 	      file->update_status = 2;
 	    }
 	  else
-	    fatal ("%s `%s'", noway, file->name);
+	    {
+	      if (file->parent == 0)
+		fatal (msg_noparent, "", file->name, "");
+	      else
+		fatal (msg_parent, "", file->name, file->parent->name, "");
+	    }
 	}
     }
   else
@@ -845,6 +870,9 @@ remake_file (file)
 	  execute_file_commands (file);
 	  return;
 	}
+      else
+	/* This tells notice_finished_file it is ok to touch the file.  */
+	file->update_status = 0;
     }
 
   /* This does the touching under -t.  */
